/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file doca_ct.h
 * @page doca ct
 * @defgroup Flow ct
 * DOCA HW connection tracking library.
 *
 * @{
 */

#ifndef DOCA_CT_H_
#define DOCA_CT_H_

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#include <doca_compat.h>
#include <doca_error.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Meta connection type
 */
enum doca_ct_meta_type {
	DOCA_CT_META_NONE,    /**< Regular payload traffic */
	DOCA_CT_META_NEW,     /**< SYN or first UDP packet */
	DOCA_CT_META_END,     /**< FIN or RST packet */
	DOCA_CT_META_UPDATE,  /**< Payload to update user action data */
};

/**
 * Meta type mask
 */
#define DOCA_CT_META_TYPE_MASK 0x3

/**
 * @brief CT packet meta data
 */
union doca_ct_meta {
	uint32_t u32; /**< CPU endian. */
	struct {
		uint32_t type : 2;     /**< Refer to doca_ct_meta_type. */
		uint32_t hairpin : 1;  /**< Subject to forward using hairpin. */
		uint32_t src : 1;      /**< Source port in multi-port E-Switch mode */
		uint32_t data : 28;    /**< Combination of reserved, zone, user action data and user data. */
	};
};

/**
 * @brief CT flags
 */
enum doca_ct_flags {
	DOCA_CT_FLAG_STATS = 1u << 0, /**< Enable counter for internal pipes */
	DOCA_CT_FLAG_WORKER_STATS = 1u << 1, /**< Enable worker counter dump */
	DOCA_CT_FLAG_NO_AGING = 1u << 2, /**< Bypass aging scan */
};

/**
 * @brief CT l3 session types
 */
enum doca_ct_session_type {
	DOCA_CT_SESSION_IPV4,		/**< IPv4 session. */
	DOCA_CT_SESSION_IPV6,		/**< IPv6 session. */
	DOCA_CT_SESSION_BOTH,		/**< Total session. */
	DOCA_CT_SESSION_MAX,		/**< Max session types. */
};

/**
 * @brief CT hash table type.
 */
enum doca_ct_hash_type {
	DOCA_CT_HASH_NONE,  /**< No hash table, besides zone, meta data bits reserved as connection ID. */
	DOCA_CT_HASH_MIXED, /**< Hardware + software hash to reserve more meta bits for user */
	DOCA_CT_HASH_RAW,   /**< Hardware raw symmetric hash key, less user meta bits */
	DOCA_CT_HASH_RAW_SYMMETRIC,   /**< Hardware raw symmetric hash function */
};

/**
 * @brief doca ct global configuration
 */
struct doca_ct_cfg {
	uint32_t nb_arm_queues; /**< number of ARM CT queues(thread). */
	uint32_t nb_arm_sessions[DOCA_CT_SESSION_MAX]; /**< number of ARM CT sessions. */
	uint32_t flags; /**< CT behavior flags */
	void *ib_dev; /**< IB verbs device context */
	void *ib_pd; /**< device protection domain */
	uint16_t tcp_timeout_s; /**< TCP timeout in second. */
	uint16_t tcp_session_del_s; /**< time to kill TCP session after RST/FIN. */
	uint16_t udp_timeout_s; /**< UDP timeout in second. */
	uint16_t aging_core; /**< CT aging thread bind to CPU core. */
	enum doca_ct_hash_type hash_type; /**< Connection hash table type. */
	uint32_t meta_user_bits; /**< User data bits ignored by worker */
	uint32_t meta_action_bits; /**< User action data bits carried by identified connection packets */
	uint32_t meta_zone_bits; /**< User zone bits carried by identified connection packets */
};

/**
 * @brief Initialize the doca ct.
 *
 * This is the global initialization function for doca ct. It
 * initializes all resources used by doca flow.
 *
 * Must be invoked first before any other function in this API.
 * this is a one time call, used for doca ct initialization and
 * global configurations.
 *
 * Must be invoked after Doca Flow initilization, before port start.
 *
 * @param cfg
 *   CT configuration.
 * @return
 *   0 on success, a negative errno value otherwise.
 */
__DOCA_EXPERIMENTAL
doca_error_t
doca_ct_init(const struct doca_ct_cfg *cfg);

/**
 * @brief Destroy the doca ct.
 *
 * Release all the resources used by doca ct.
 *
 * Must be invoked before doca flow detroy.
 */
__DOCA_EXPERIMENTAL
void
doca_ct_destroy(void);

/**
 * @brief Get meta zone data.
 *
 * @param meta
 *   CT meta.
 * @return
 *   Zone value.
 */
__DOCA_EXPERIMENTAL
uint32_t
doca_ct_meta_get_zone(uint32_t meta);

/**
 * @brief Set meta zone data applies to identified connection packets.
 *
 * @param meta
 *   CT meta.
 * @param zone
 *   Zone value.
 */
__DOCA_EXPERIMENTAL
void
doca_ct_meta_set_zone(uint32_t *meta, uint32_t zone);

/**
 * @brief Set meta action data applies to identified connection packets.
 *
 * @param meta
 *   CT meta.
 * @param action_data
 *   Action data.
 */
__DOCA_EXPERIMENTAL
void
doca_ct_meta_set_action(uint32_t *meta, uint32_t action_data);

/**
 * @brief Set user data in meta data field.
 *
 * User data is ignored by worker, can't be carried with identified conneciton packets.
 * @param meta
 *   CT meta.
 * @param user_data
 *   User data value.
 */
__DOCA_EXPERIMENTAL
void
doca_ct_meta_set_user(uint32_t *meta, uint32_t user_data);

/**
 * @brief Get zone data bit offset in meta data field.
 *
 * @return
 *   Zone data bit offset.
 */
__DOCA_EXPERIMENTAL
uint32_t
doca_ct_meta_get_zone_offset(void);

/**
 * @brief Get action data bit offset in meta data field.
 *
 * @return
 *   Action data bit offset.
 */
__DOCA_EXPERIMENTAL
uint32_t
doca_ct_meta_get_action_offset(void);

/**
 * @brief Get User data bit offset in meta data field.
 *
 * @return
 *   User data bit offset.
 */
__DOCA_EXPERIMENTAL
uint32_t
doca_ct_meta_get_user_offset(void);

#ifdef __cplusplus
} /* extern "C" */
#endif

/** @} */

#endif /* DOCA_CT_H_ */
